# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

from collections import defaultdict
import numpy as np

from wittgenstein.base_functions import truncstr
from wittgenstein.utils import rnd


class BinTransformer:
    def __init__(self, n_discretize_bins=10, names_precision=2, verbosity=0):
        self.n_discretize_bins = n_discretize_bins
        self.names_precision = names_precision
        self.verbosity = verbosity
        self.bins_ = None

    def __str__(self):
        return str(self.bins_)

    __repr__ = __str__

    def __bool__(self):
        return not not self.bins_

    def isempty(self):
        return not self.bins_ is None and not self.bins_

    def fit_or_fittransform_(self, df, ignore_feats=[]):
        """Transform df using pre-fit bins, or, if unfit, fit self and transform df"""

        # Binning has already been fit
        if self.bins_:
            return self.transform(df)

        # Binning disabled
        elif not self.n_discretize_bins:
            return df

        # Binning enabled, and binner needs to be fit
        else:
            self.fit(df, ignore_feats=ignore_feats)
            df, bins = self.transform(df, ignore_feats=ignore_feats)
            self.bins = bins
            return df

    def fit_transform(self, df, ignore_feats=[]):
        self.fit(df, ignore_feats=ignore_feats)
        return self.transform(df)

    def transform(self, df, ignore_feats=[]):
        """Return df with any continuous features binned, and the bin_transformer or None depending on whether binning occurs."""

        if self.n_discretize_bins is None:
            return df

        if self.bins_ == {}:
            return df

        isbinned = False
        continuous_feats = find_continuous_feats(df, ignore_feats=ignore_feats)
        if self.n_discretize_bins:
            if continuous_feats:
                if self.verbosity == 1:
                    print(f"binning data...\n")
                elif self.verbosity >= 2:
                    print(f"binning features {continuous_feats}...")
                binned_df = df.copy()
                bin_transformer = fit_bins(
                    binned_df, output=False, ignore_feats=ignore_feats,
                )
                binned_df = bin_transform(binned_df, bin_transformer)
                isbinned = True
        else:
            n_unique_values = sum(
                [len(u) for u in [df[f].unique() for f in continuous_feats]]
            )
            warning_str = f"There are {len(continuous_feats)} features to be treated as continuous: {continuous_feats}. \n Treating {n_unique_values} numeric values as nominal or discrete. To auto-discretize features, assign a value to parameter 'n_discretize_bins.'"
            _warn(warning_str, RuntimeWarning, filename="base", funcname="transform")
        if isbinned:
            self.bins_ = bin_transformer
            return binned_df, bin_transformer
        else:
            return df

    def find_continuous_feats(self, df, ignore_feats=[]):
        """Return names of df features that seem to be continuous."""

        if not self.n_discretize_bins:
            return []

        # Find numeric features
        cont_feats = df.select_dtypes(np.number).columns

        # Remove discrete features
        cont_feats = [
            f for f in cont_feats if len(df[f].unique()) > self.n_discretize_bins
        ]

        # Remove ignore features
        cont_feats = [f for f in cont_feats if f not in ignore_feats]

        return cont_feats

    def fit(self, df, output=False, ignore_feats=[]):
        """
        Returns a dict defining fits for numerical features
        A fit is an ordered list of tuples defining each bin's range (min is exclusive; max is inclusive)

        Returned dict allows for fitting to training data and applying the same fit to test data
        to avoid information leak.
        """

        def _fit_feat(df, feat):
            """Return list of tuples defining bin ranges for a numerical feature using simple linear search"""

            if len(df) == 0:
                return []

            n_discretize_bins = min(
                self.n_discretize_bins, len(df[feat].unique())
            )  # In case there are fewer unique values than n_discretize_bins
            bin_size = len(df) // n_discretize_bins
            sorted_df = df.sort_values(by=[feat])
            sorted_values = sorted_df[feat].tolist()

            sizes = []  # for verbosity output
            if self.verbosity >= 3:
                print(
                    f"{feat}: fitting {len(df[feat].unique())} unique vals into {n_discretize_bins} bins"
                )

            bin_ranges = []  # result
            bin_num = 0  # current bin number

            ceil_i = -1  # current bin ceiling index
            ceil_val = None  # current bin upper bound

            floor_i = 0  # current bin start index
            floor_val = sorted_df.iloc[0][feat]  # current bin floor value

            prev_finish_val = None  # prev bin upper bound
            while bin_num < n_discretize_bins and floor_i < len(sorted_values):
                # jump to tentative ceiling index
                ceil_i = min(floor_i + bin_size, len(sorted_df) - 1)
                ceil_val = sorted_df.iloc[ceil_i][feat]

                # increment ceiling index until encounter a new value to ensure next bin size is correct
                while (
                    ceil_i < len(sorted_df) - 1  # not last bin
                    and sorted_df.iloc[ceil_i][feat]
                    == ceil_val  # keep looking for a new value
                ):
                    ceil_i += 1

                # found ceiling index. update values
                if self.verbosity >= 4:
                    sizes.append(ceil_i - floor_i)
                    print(
                        f"bin #{bin_num}, floor idx {floor_i} value: {sorted_df.iloc[floor_i][feat]}, ceiling idx {ceil_i} value: {sorted_df.iloc[ceil_i][feat]}"
                    )
                bin_range = (floor_val, ceil_val)
                bin_ranges.append(bin_range)

                # update for next bin
                floor_i = ceil_i + 1
                floor_val = ceil_val
                bin_num += 1

            # Guarantee min and max values
            bin_ranges[0] = (sorted_df.iloc[0][feat], bin_ranges[0][1])
            bin_ranges[-1] = (bin_ranges[-1][0], sorted_df.iloc[-1][feat])

            if self.verbosity >= 4:
                print(
                    f"-bin sizes {sizes}; dataVMR={rnd(np.var(df[feat])/np.mean(df[feat]))}, binVMR={rnd(np.var(sizes)/np.mean(sizes))}"
                )  # , axis=None, dtype=None, out=None, ddof=0)})
            return bin_ranges

        # Create dict to store fit definitions for each feature
        fit_dict = {}
        feats_to_fit = self.find_continuous_feats(df, ignore_feats=ignore_feats)
        if feats_to_fit:
            if self.verbosity == 1:
                print(f"discretizing {len(feats_to_fit)} features")
            if self.verbosity == 2:
                print(f"fitting bins for {len(feats_to_fit)} features: {feats_to_fit}")
            if self.verbosity >= 2:
                print()

        # Collect fits in dict
        for feat in feats_to_fit:
            fit = _fit_feat(df, feat)
            fit_dict[feat] = fit
        self.bins_ = fit_dict

    def transform(self, df):
        """
        Uses a pre-collected dictionary of fits to transform df features into bins.
        Returns the fit df rather than modifying inplace.
        """

        if self.bins_ is None:
            return df

        # Replace each feature with bin transformations
        for feat, bin_fit_list in self.bins_.items():
            if feat in df.columns:
                df[feat] = df[feat].map(
                    lambda x: self._transform_value(x, bin_fit_list)
                )
        return df

    def _transform_value(self, value, bin_fit_list):
        """Return bin string name for a given numerical value. Assumes bin_fit_list is ordered."""
        min_val, min_bin = bin_fit_list[0][0], bin_fit_list[0]
        max_val, max_bin = bin_fit_list[-1][1], bin_fit_list[-1]
        for bin_fit in bin_fit_list:
            if value <= bin_fit[1]:
                start_name = (
                    str(round(bin_fit[0], self.names_precision))
                    if self.names_precision
                    else str(int(bin_fit[0]))
                )
                finish_name = (
                    str(round(bin_fit[1], self.names_precision))
                    if self.names_precision
                    else str(int(bin_fit[1]))
                )
                bin_name = "-".join([start_name, finish_name])
                return bin_name
        if value <= min_val:
            return min_bin
        elif value >= max_val:
            return max_bin
        else:
            raise ValueError("No bin found for value", value)

    def _try_rename_features(self, df, class_feat, feature_names):
        """Rename df columns according to user request."""
        # Rename if same number of features
        df_columns = [col for col in df.columns.tolist() if col != class_feat]
        if len(df_columns) == len(feature_names):
            col_replacements_dict = {
                old: new for old, new in zip(df_columns, feature_names)
            }
            df = df.rename(columns=col_replacements_dict)
            return df
        # Wrong number of feature names
        else:
            return None

    def construct_from_ruleset(self, ruleset):
        MIN_N_DISCRETIZED_BINS = 10

        bt = BinTransformer()
        bt.bins_ = self._bin_prediscretized_features(ruleset)
        bt.n_discretize_bins = (
            max(
                (MIN_N_DISCRETIZED_BINS, max([len(bins) for bins in bt.bins_.values()]))
            )
            if bt.bins_
            else MIN_N_DISCRETIZED_BINS
        )
        bt.names_precision = self._max_dec_precision(bt.bins_)
        return bt

    def _bin_prediscretized_features(self, ruleset):
        def is_valid_decimal(s):
            try:
                float(s)
            except:
                return False
            return True

        def find_floor_ceil(value):
            """id min, max separated by a dash. Return None if invalid pattern."""
            split_idx = 0
            for i, char in enumerate(value):
                # Found a possible split and it's not the first number's minus sign
                if char == "-" and i != 0:
                    if split_idx is not None and not split_idx:
                        split_idx = i
                    # Found a - after the split, and it's not the minus of a negative number
                    elif i > split_idx + 1:
                        return None

            floor = value[:split_idx]
            ceil = value[split_idx + 1 :]
            if is_valid_decimal(floor) and is_valid_decimal(ceil):
                return (floor, ceil)
            else:
                return None

        # _bin_prediscretized_features
        discrete = defaultdict(list)
        for cond in ruleset.get_conds():
            floor_ceil = find_floor_ceil(cond.val)
            if floor_ceil:
                discrete[cond.feature].append(floor_ceil)
        for feat, ranges in discrete.items():
            ranges.sort(key=lambda x: float(x[0]))
        return dict(discrete)

    def _max_dec_precision(self, bins_dict):
        def dec_precision(value):
            try:
                return len(value) - value.index(".") - 1
            except:
                return 0

        max_prec = 0
        for bins in bins_dict.values():
            for bin_ in bins:
                for value in bin_:
                    cur_prec = dec_precision(value)
                    if cur_prec > max_prec:
                        max_prec = cur_prec
        return max_prec



# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

from wittgenstein.base_functions import score_accuracy
from wittgenstein.ripper import RIPPER
from wittgenstein import utils


def interpret_model(
    X,
    model,
    interpreter=RIPPER(),
    model_predict_function=None,
    score_function=score_accuracy,
):
    """Interpret another model using a wittgenstein interpreter as global surrogate.

        Parameters
        ----------
        model :
            trained classifier, e.g. sklearn, keras, pytorch, etc.
        X : DataFrame, numpy array, or other iterable
            Dataset upon which to interpret model's predictions.
        interpreter : IREP or RIPPER object, default=RIPPER()
            wittgenstein classifier to perform interpretation.
        model_predict_function : function, default=None
            if
        score_function : function, default=score_accuracy
            scoring function to evaluate how dutifully interpreter interpreted the model.

        Return
        ------
        tuple :
            interpreter fit to model,
            scoring of how dutifully interpreter interpreted the model on training data
    """
    model_preds = utils.try_np_tonum(
        model_predict(X, model, model_predict_function=model_predict_function)
    )
    interpreter.fit(trainset=X, y=model_preds)
    interpreter.base_model = model
    return interpreter.ruleset_


def score_fidelity(
    X,
    interpreter,
    model=None,
    model_preds=None,
    model_predict_function=None,
    score_function=score_accuracy,
):
    """Score how faithfully interpreter represents model.

    Parameters
    ----------
    X : DataFrame, numpy array, or other iterable
        Test dataset with which to score the model.
    interpreter : IREP or RIPPER object, default=RIPPER()
        wittgenstein classifier to perform interpretation.
    model : trained sklearn, keras, pytorch, or wittgenstein, etc. classifier, default=None
        either model or model_preds are required
    model_preds : iterable
        model predictions on X, default=None
    model_predict_function : function, default=None
        model's prediction function. If None, will attempt to figure it out.
    score_function : function or iterable of functions, default=score_accuracy
        criteria to use for scoring fidelity

    Returns
    -------
    score or list of scores"""

    if model is None and model_preds is None:
        raise ValueError(f"score_fidelity: You must pass a model or model predictions")
    elif model_preds is None:
        model_preds = utils.try_np_tonum(
            model_predict(X, model, model_predict_function=model_predict_function)
        )
    if not hasattr(score_function, "__iter__"):
        return interpreter.score(X, model_preds, score_function)
    else:
        return [interpreter.score(X, model_preds, func) for func in score_function]


def model_predict(X, model, model_predict_function=None):
    """Attempt to make predictions using model API"""

    if not model_predict_function:
        if _inpackage(model, "sklearn"):
            return _sklearn_predict(X, model)
        elif _inpackage(model, "tensorflow") or _inpackage(model, "keras"):
            return _keras_predict(X, model)
        elif _inpackage(model, "torch"):
            return _torch_predict(X, model)
        elif _inpackage(model, "wittgenstein"):
            return _wittgenstein_predict(X, model)
        else:
            return model.predict(X)
    else:
        return model_predict_function(X, model)


def _sklearn_predict(X, model):
    return model.predict(X)


def _keras_predict(X, model):
    return (model.predict(X) > 0.5).flatten()


def _torch_predict(X, model):
    return model(X)


def _wittgenstein_predict(X, model):
    return model.predict(X)


def _inpackage(model, str_):
    return str_ in str(type(model))


def _score_model(
    X, y, model, score_function=score_accuracy, model_predict_function=None
):
    model_preds = utils.try_np_tonum(
        model_predict(X, model=model, model_predict_function=model_predict_function)
    )
    return score_function(model_preds, y)



"""
Implementation of incremental reduced error pruning (IREP*) algorithm for growing classification rulesets.
See https://www.let.rug.nl/nerbonne/teach/learning/cohen95fast.pdf
"""

# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import copy
import numpy as np
import random

import pandas as pd

from wittgenstein import utils, base, base_functions, preprocess
from .catnap import CatNap
from .check import _check_param_deprecation, _check_df_allpos_allneg
from .abstract_ruleset_classifier import AbstractRulesetClassifier
from .base import Cond, Rule, Ruleset, asruleset
from .base_functions import score_accuracy, stop_early


class IREP(AbstractRulesetClassifier):
    """ Class for generating ruleset classification models. """

    def __init__(
        self,
        prune_size=0.33,
        n_discretize_bins=10,
        max_rules=None,
        max_rule_conds=None,
        max_total_conds=None,
        alpha=1.0,
        random_state=None,
        verbosity=0,
    ):
        """Create an IREP classifier.

        Parameters
        ----------
        prune_size : int, default=.33
            Proportion of training set to be used for pruning. Set to None to skip pruning (not recommended).
        n_discretize_bins : int, default=10
            Fit apparent numeric attributes into a maximum of n_discretize_bins discrete bins, inclusive on upper part of range. Pass None to disable auto-discretization.

        Limits for early-stopping. Intended for enhancing model interpretability and limiting training time on noisy datasets. Not specifically intended for use as a hyperparameter, since pruning already occurs during training, though it is certainly possible that tuning could improve model performance.
        max_rules : int, default=None
            Maximum number of rules.
        max_rule_conds : int, default=None
            Maximum number of conds per rule.
        max_total_conds : int, default=None
            Maximum number of total conds in entire ruleset.

        random_state : int, default=None
            Random seed for repeatable results.
        verbosity : int, default=0
            Output progress, model development, and/or computation. Each level includes the information belonging to lower-value levels.
               1: Show results of each major phase
               2: Show Ruleset grow/optimization steps
               3: Show Ruleset grow/optimization calculations
               4: Show Rule grow/prune steps
               5: Show Rule grow/prune calculations

        """
        AbstractRulesetClassifier.__init__(
            self,
            algorithm_name="IREP",
            prune_size=prune_size,
            n_discretize_bins=n_discretize_bins,
            max_rules=max_rules,
            max_rule_conds=max_rule_conds,
            max_total_conds=max_total_conds,
            alpha=alpha,
            random_state=random_state,
            verbosity=verbosity,
        )

    def __str__(self):
        """Returns string representation."""
        return super().__str__()

    def out_model(self):
        """Prints trained Ruleset model line-by-line: V represents 'or'; ^ represents 'and'."""
        super().out_model()

    def fit(
        self,
        trainset,
        y=None,
        class_feat=None,
        pos_class=None,
        feature_names=None,
        initial_model=None,
        cn_optimize=True,
        **kwargs,
    ):
        """Fit a Ruleset model.

        Parameters
        ----------
        trainset : DataFrame, numpy array, or other iterable
            Training dataset. Optional whether to include or exclude class labels column.
        y : iterable of str, int, bool
            Class labels corresponding to trainset rows. Use if class labels aren't included in trainset.
        class_feat: str, int
            Column name or index of class feature. Use if class feature is still in trainset.
        pos_class : str, optional for boolean target, default=1 or True
            Name of positive class.
        feature_names : list<str>, optional, default=None
            Specify feature names. If None, feature names default to column names for a DataFrame, or indices in the case of indexed iterables such as an array or list.
        initial_model : Ruleset, str, IREP or RIPPER, default=None
            Preexisting model from which to begin training. See also 'init_ruleset'.
        cn_optimize : bool, default=True
            Use algorithmic speed optimization.

        **kwargs
        --------
        The following parameters are moving to the RIPPER constructor (__init__) function. For the time-being, both the constructor and fit functions will accept them, but passing them here using .fit will be deprecated:

        prune_size : float, default=.33
            Proportion of training set to be used for pruning.
        n_discretize_bins : int, default=10
            Fit apparent numeric attributes into a maximum of n_discretize_bins discrete bins, inclusive on upper part of range. Pass None to disable auto-discretization.
        random_state : int, default=None
            Random seed for repeatable results.

        Limits for early-stopping. Intended for enhancing model interpretability and limiting training time on noisy datasets. Not specifically intended for use as a hyperparameter, since pruning already occurs during training, though it is certainly possible that tuning could improve model performance.
        max_rules : int, default=None
            Maximum number of rules.
        max_rule_conds : int, default=None
            Maximum number of conds per rule.
        max_total_conds : int, default=None
            Maximum number of total conds in entire ruleset.

        verbosity : int, default=0
            Output progress, model development, and/or computation. Each level includes the information belonging to lower-value levels.
               1: Show results of each major phase
               2: Show Ruleset grow/optimization steps
               3: Show Ruleset grow/optimization calculations
               4: Show Rule grow/prune steps
               5: Show Rule grow/prune calculations
        """

        # SETUP
        self.ruleset_ = Ruleset() if not initial_model else asruleset(initial_model)

        if self.verbosity >= 1:
            print("initialize model")
            print(self.ruleset_)

        # Handle any hyperparam deprecation
        self._set_deprecated_fit_params(kwargs)

        # Preprocess training data
        preprocess_params = {
            "trainset": trainset,
            "y": y,
            "class_feat": class_feat,
            "pos_class": pos_class,
            "feature_names": feature_names,
            "n_discretize_bins": self.n_discretize_bins,
            "verbosity": self.verbosity,
        }
        (
            df,
            self.class_feat,
            self.pos_class,
            self.bin_transformer_,
        ) = preprocess.preprocess_training_data(preprocess_params)

        # Create CatNap
        # possible minor speedup if pass cond_subset of only pos_class conds?
        if cn_optimize:
            self.cn = CatNap(
                df,
                feat_subset=None,
                cond_subset=None,
                class_feat=self.class_feat,
                pos_class=None,
            )

        # Split df into pos, neg classes
        pos_df, neg_df = base_functions.pos_neg_split(
            df, self.class_feat, self.pos_class
        )
        pos_df = pos_df.drop(self.class_feat, axis=1)
        neg_df = neg_df.drop(self.class_feat, axis=1)

        # Warnings
        _check_df_allpos_allneg(df, self.class_feat, self.pos_class, 'irep', 'fit')

        # TRAINING
        if self.verbosity >= 1:
            print("\ntraining Ruleset...")
        if not cn_optimize:
            self.ruleset_ = self._grow_ruleset(
                pos_df, neg_df, initial_model=initial_model
            )
        else:
            self.ruleset_ = self._grow_ruleset_cn(
                pos_df, neg_df, initial_model=initial_model
            )

        if self.verbosity >= 1:
            print("\nGREW RULESET:")
            self.ruleset_.out_pretty()

        # Issue warning if Ruleset is universal or empty
        self.ruleset_._check_allpos_allneg(warn=True, warnstack=[("irep", "fit")])

        # Set selected and trainset features
        self.selected_features_ = self.ruleset_.get_selected_features()
        self.trainset_features_ = df.drop(self.class_feat, axis=1).columns.tolist()

        # FIT PROBAS
        self.recalibrate_proba(
            df, min_samples=1, require_min_samples=False, discretize=False
        )

        # CLEANUP
        self.classes_ = np.array([0, 1])

        # Remove any duplicates and trim
        self.ruleset_.rules = utils.remove_duplicates(self.ruleset_.rules)
        self.ruleset_.trim_conds(max_total_conds=self.max_total_conds)
        if cn_optimize:
            del self.cn

    def _grow_ruleset(self, pos_df, neg_df, initial_model=None):
        """Grow a Ruleset with (optional) pruning."""

        ruleset = self._ruleset_frommodel(initial_model)
        ruleset._update_possible_conds(pos_df, neg_df)

        if self.verbosity >= 2:
            print("growing ruleset...")
            print(f"initial model: {ruleset}")
            print()

        prune_size = (
            self.prune_size if self.prune_size is not None else 0
        )  # If not pruning, use all the data for growing
        pos_remaining = pos_df.copy()
        neg_remaining = neg_df.copy()
        self.rules = []

        # Stop adding disjunctions if there are no more positive examples to cover
        while len(pos_remaining) > 0:

            # If applicable, check for user-specified early stopping
            if stop_early(ruleset, self.max_rules, self.max_total_conds):
                break

            # Grow-prune split remaining uncovered examples (if applicable)
            pos_growset, pos_pruneset = base_functions.df_shuffled_split(
                pos_remaining, (1 - prune_size), random_state=self.random_state
            )
            neg_growset, neg_pruneset = base_functions.df_shuffled_split(
                neg_remaining, (1 - prune_size), random_state=self.random_state
            )
            if self.verbosity >= 2:
                print(
                    f"pos_growset {len(pos_growset)} pos_pruneset {len(pos_pruneset)}"
                )
                print(
                    f"neg_growset {len(neg_growset)} neg_pruneset {len(neg_pruneset)}"
                )
                if not prune_size:
                    print(f"(pruning is turned off)")

            # Grow Rule
            grown_rule = base_functions.grow_rule(
                pos_growset,
                neg_growset,
                ruleset.possible_conds,
                max_rule_conds=self.max_rule_conds,
                verbosity=self.verbosity,
            )

            # If not pruning, add Rule to Ruleset and drop only the covered positive examples
            if not prune_size:
                ruleset.add(grown_rule)
                if self.verbosity >= 2:
                    print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                    print()
                rule_covers_pos = grown_rule.covers(pos_remaining)
                pos_remaining = pos_remaining.drop(rule_covers_pos.index, axis=0)
                if self.verbosity >= 3:
                    print(
                        f"examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg"
                    )
                    print()

            # If pruning, prune Rule, assess if it's time to stop, and drop all covered examples
            else:
                pruned_rule = base_functions.prune_rule(
                    grown_rule,
                    _IREP_prune_metric,
                    pos_pruneset,
                    neg_pruneset,
                    verbosity=self.verbosity,
                )

                # Stop if the Rule is bad
                prune_precision = base_functions.precision(
                    pruned_rule, pos_pruneset, neg_pruneset
                )
                if not prune_precision or prune_precision < 0.50:
                    break
                # Otherwise, add new Rule, remove covered examples, and continue
                else:
                    ruleset.add(pruned_rule)
                    if self.verbosity >= 2:
                        print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                        print()
                    pos_remaining, neg_remaining = base_functions.rm_covered(
                        pruned_rule, pos_remaining, neg_remaining
                    )
                    if self.verbosity >= 3:
                        print(
                            f"examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg"
                        )
                        print()

        # Return new ruleset
        return ruleset

    def _grow_ruleset_cn(self, pos_df, neg_df, initial_model=None):
        """Grow a Ruleset with (optional) pruning."""

        ruleset = self._ruleset_frommodel(initial_model)
        ruleset.possible_conds = self.cn.conds

        if self.verbosity >= 2:
            print("growing ruleset...")
            print(initial_model)
            print(f"initial model: {ruleset}")
            print()

        prune_size = (
            self.prune_size if self.prune_size is not None else 0
        )  # If not pruning, use all the data for growing
        pos_remaining_idx = set(pos_df.index.tolist())
        neg_remaining_idx = set(neg_df.index.tolist())

        # Stop adding disjunctions if there are no more positive examples to cover
        while len(pos_remaining_idx) > 0:

            # If applicable, check for user-specified early stopping
            if stop_early(ruleset, self.max_rules, self.max_total_conds):
                break

            # Grow-prune split remaining uncovered examples (if applicable)
            pos_growset_idx, pos_pruneset_idx = base_functions.random_split(
                pos_remaining_idx,
                (1 - prune_size),
                res_type=set,
                random_state=self.random_state,
            )
            neg_growset_idx, neg_pruneset_idx = base_functions.random_split(
                neg_remaining_idx,
                (1 - prune_size),
                res_type=set,
                random_state=self.random_state,
            )

            if self.verbosity >= 2:
                print(
                    f"pos_growset {len(pos_growset_idx)} pos_pruneset {len(pos_pruneset_idx)}"
                )
                print(
                    f"neg_growset {len(neg_growset_idx)} neg_pruneset {len(neg_pruneset_idx)}"
                )
                if not prune_size:
                    print(f"(pruning is turned off)")

            # Grow Rule
            grown_rule = base_functions.grow_rule_cn(
                self.cn,
                pos_growset_idx,
                neg_growset_idx,
                initial_rule=Rule(),
                max_rule_conds=self.max_rule_conds,
                verbosity=self.verbosity,
            )

            # If not pruning, add Rule to Ruleset and drop only the covered positive examples
            if not prune_size:
                ruleset.add(grown_rule)
                if self.verbosity >= 2:
                    print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                    print()
                pos_remaining_idx = pos_remaining_idx - self.cn.rule_covers(
                    grown_rule, pos_remaining_idx
                )
                if self.verbosity >= 3:
                    print(
                        f"examples remaining: {len(pos_remaining_idx)} pos, {len(neg_remaining_idx)} neg"
                    )
                    print()

            # If pruning, prune Rule, assess if it's time to stop, and drop all covered examples
            else:
                pruned_rule = base_functions.prune_rule_cn(
                    self.cn,
                    grown_rule,
                    _IREP_prune_metric_cn,
                    pos_pruneset_idx,
                    neg_pruneset_idx,
                    verbosity=self.verbosity,
                )

                # Stop if the Rule is bad
                prune_precision = base_functions.rule_precision_cn(
                    self.cn, pruned_rule, pos_pruneset_idx, neg_pruneset_idx
                )
                print('prune_precision for rule', pruned_rule, prune_precision)
                if not prune_precision or prune_precision < 0.50:
                    break
                # Otherwise, add new Rule, remove covered examples, and continue
                else:
                    ruleset.add(pruned_rule)
                    if self.verbosity >= 2:
                        print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                        print()
                    # Remove ruleset-covered rules from training
                    (
                        pos_remaining_idx,
                        neg_remaining_idx,
                    ) = base_functions.rm_rule_covers_cn(
                        self.cn, pruned_rule, pos_remaining_idx, neg_remaining_idx
                    )
                    if self.verbosity >= 3:
                        print(
                            f"examples remaining: {len(pos_remaining_idx)} pos, {len(neg_remaining_idx)} neg"
                        )
                        print()

        # Return new ruleset
        return ruleset


##### Metrics #####


def _IREP_prune_metric(self, pos_pruneset, neg_pruneset):
    """Returns the prune value of a candidate Rule."""

    P = len(pos_pruneset)
    N = len(neg_pruneset)
    p = self.num_covered(pos_pruneset)
    n = self.num_covered(neg_pruneset)
    return (p + (N - n)) / (P + N)


def _IREP_prune_metric_cn(cn, rule, pos_idxs, neg_idxs):
    """Returns the prune value of a candidate Rule."""

    P = len(pos_idxs)
    N = len(neg_idxs)
    p = len(cn.rule_covers(rule, pos_idxs))
    n = len(cn.rule_covers(rule, neg_idxs))
    return (p + (N - n)) / (P + N)



import numpy as np
from numpy.random import seed

import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.metrics import precision_score, recall_score, f1_score
from tensorflow.keras import Sequential
from keras.layers import Dense

from wittgenstein.irep import IREP
from wittgenstein.ripper import RIPPER
from wittgenstein.interpret import interpret_model, score_fidelity

inpath = ""
df = pd.read_csv(inpath + "credit.csv")
class_feat = "Class"

a2_mean = np.mean([float(val) for val in df["A2"].tolist() if val != "?"])
df["A2"] = df["A2"].map(lambda x: float(x) if x != "?" else a2_mean)
df["A14"] = df["A14"].map(lambda x: float(x) if x != "?" else a2_mean)
df["A15"] = df["A15"].map(lambda x: float(x) if x != "?" else a2_mean)

cat_feats = [
    col
    for col in df.columns
    if (col not in df.select_dtypes("number").columns and col != class_feat)
]
df = pd.get_dummies(df, columns=cat_feats)

X, y = df.drop("Class", axis=1), df["Class"]
y = y.map(lambda x: 1 if x == "+" else 0)

X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)


def test_interpret_keras():
    seed(1)
    wine_df = pd.read_csv(inpath + "wine.csv")
    X_wine, y_wine = wine_df.drop("Class", axis=1), wine_df["Class"]
    y_wine = y_wine.map(lambda x: True if x==1 else False)
    X_train_wine, X_test_wine, y_train_wine, y_test_wine = train_test_split(X, y, random_state=42)

    model = Sequential()
    model.add(Dense(60, input_dim=13, activation='relu'))
    model.add(Dense(30, activation='relu'))
    model.add(Dense(1, activation='sigmoid'))
    model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
    model.fit(
        X_train_wine,
        y_train_wine,
        batch_size=1,
        epochs=10,
    )

    base_model = model
    irep = IREP(random_state=42)
    interpret_model(model=base_model, X=X_test_wine, interpreter=irep)
    assert (irep.base_model) and (not irep.ruleset_.isuniversal()) and (not irep.ruleset_.isnull())
    irep.predict(X_test)

    rip = RIPPER(random_state=42)
    interpret_model(model=base_model, X=X_test_wine, interpreter=rip)
    assert (rip.base_model) and (not rip.ruleset_.isuniversal()) and (not rip.ruleset_.isnull())
    rip.predict(X_test)


def test_interpret_svc():
    svc = SVC(random_state=42)
    svc.fit(X_train, y_train)

    irep = IREP(random_state=42)
    interpret_model(model=svc, X=X_test, interpreter=irep)
    assert sum(irep.predict(X_test)) == 18

    rip = RIPPER(random_state=42)
    interpret_model(model=svc, X=X_test, interpreter=rip)
    assert sum(rip.predict(X_test)) == 22


def test_score_fidelity():
    svc = SVC(kernel='rbf', random_state=42)
    svc.fit(X_train, y_train)

    irep = IREP(random_state=42)
    interpret_model(model=svc, X=X_test, interpreter=irep)
    score = score_fidelity(
        X_test,
        irep,
        model=svc,
        score_function=[precision_score, recall_score, f1_score])
    assert all([sc > 0.5 for sc in score])

    rip = RIPPER(random_state=42)
    interpret_model(model=svc, X=X_test, interpreter=rip)
    score = score_fidelity(
        X_test,
        rip,
        model=svc,
        score_function=[precision_score, recall_score, f1_score])
    assert all([sc > 0.5 for sc in score])



# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

from copy import deepcopy
from collections import defaultdict
import numpy as np
import pandas as pd

from wittgenstein.base_functions import truncstr
from wittgenstein.utils import rnd
from wittgenstein.check import _warn, is_valid_decimal

class BinTransformer:
    def __init__(self, n_discretize_bins=10, names_precision=2, verbosity=0):
        self.n_discretize_bins = n_discretize_bins
        self.names_precision = names_precision
        self.verbosity = verbosity
        self.bins_ = None

    def __str__(self):
        return str(self.bins_)

    __repr__ = __str__

    def __bool__(self):
        return not not self.bins_

    def isempty(self):
        return not self.bins_ is None and not self.bins_

    def fit_or_fittransform_(self, df, ignore_feats=[]):
        """Transform df using pre-fit bins, or, if unfit, fit self and transform df"""

        # Binning has already been fit
        if self.bins_:
            return self.transform(df)

        # Binning disabled
        elif not self.n_discretize_bins:
            return df

        # Binning enabled, and binner needs to be fit
        else:
            self.fit(df, ignore_feats=ignore_feats)
            df, bins = self.transform(df, ignore_feats=ignore_feats)
            self.bins = bins
            return df

    def fit_transform(self, df, ignore_feats=[]):
        self.fit(df, ignore_feats=ignore_feats)
        return self.transform(df)

    def fit(self, df, output=False, ignore_feats=[]):
        """
        Returns a dict defining fits for numerical features
        A fit is an ordered list of tuples defining each bin's range (min is exclusive; max is inclusive)

        Returned dict allows for fitting to training data and applying the same fit to test data
        to avoid information leak.
        """

        def _fit_feat(df, feat):
            """Return list of tuples defining bin ranges for a numerical feature using simple linear search"""

            if len(df) == 0:
                return []

            n_discretize_bins = min(self.n_discretize_bins, len(df[feat].unique()))

            # Collect intervals
            bins = pd.qcut(
                df[feat],
                q=self.n_discretize_bins,
                precision=self.names_precision,
                duplicates="drop",
            )
            if (
                len(bins.unique()) < 2
            ):  # qcut can behave weirdly in heavily-skewed distributions
                bins = pd.cut(
                    df[feat],
                    bins=self.n_discretize_bins,
                    precision=self.names_precision,
                    duplicates="drop",
                )

            # Drop empty bins and duplicate intervals to create bins
            bin_counts = bins.value_counts()
            bins = bin_counts[bin_counts > 0].index
            bins = sorted(bins.unique())

            # Extend min/max to -inf, +inf to capture any ranges not present in training set
            bins[0] = pd.Interval(float("-inf"), bins[0].right)
            bins[-1] = pd.Interval(bins[-1].left, float("inf"))
            bins = self._intervals_to_strs(bins)

            if self.verbosity >= 3:
                print(
                    f"{feat}: fit {len(df[feat].unique())} unique vals into {len(bins)} bins"
                )
            return bins

        # Begin fitting
        feats_to_fit = self.find_continuous_feats(df, ignore_feats=ignore_feats)

        if feats_to_fit:
            if self.verbosity == 1:
                print(f"discretizing {len(feats_to_fit)} features")
            elif self.verbosity == 2:
                print(f"discretizing {len(feats_to_fit)} features: {feats_to_fit}\n")

        self.bins_ = {}
        for feat in feats_to_fit:
            self.bins_[feat] = _fit_feat(df, feat)
        return self.bins_

    def transform(self, df):
        """Transform DataFrame using fit bins."""

        def _transform_feat(df, feat):

            if self.bins_ is None:
                return df

            res = deepcopy(df[feat])
            bins = self._strs_to_intervals(self.bins_[feat], feat)
            res = pd.cut(df[feat], bins=pd.IntervalIndex(bins))
            res = res.map(
                lambda x: {i: s for i, s in zip(bins, self.bins_[feat])}.get(x)
            )
            return res

        # Exclude any feats already transformed into valid intervals
        already_transformed_feats = self._find_transformed(df, raise_invalid=True)

        res = df.copy()
        for feat in self.bins_.keys():
            if feat in res.columns and feat not in already_transformed_feats:
                res[feat] = _transform_feat(res, feat)
        return res

    def find_continuous_feats(self, df, ignore_feats=[]):
        """Return names of df features that seem to be continuous."""

        if not self.n_discretize_bins:
            return []

        # Find numeric features
        cont_feats = df.select_dtypes(np.number).columns

        # Remove discrete features
        cont_feats = [
            f for f in cont_feats if len(df[f].unique()) > self.n_discretize_bins
        ]

        # Remove ignore features
        cont_feats = [f for f in cont_feats if f not in ignore_feats]

        return cont_feats

    def _strs_to_intervals(self, strs, feat):
        return [self._str_to_interval(s, feat) for s in strs]

    def _str_to_interval(self, s, feat):
        floor, ceil = self._str_to_floor_ceil(s, feat)
        return pd.Interval(floor, ceil)

    def _intervals_to_strs(self, intervals):
        """Replace a list of intervals with their string representation."""
        return [self._interval_to_str(interval) for interval in intervals]

    def _interval_to_str(self, interval):
        if interval.left == float("-inf"):
            return f"<{interval.right}"
        elif interval.right == float("inf"):
            return f">{interval.left}"
        else:
            return f"{interval.left} - {interval.right}"

    def _str_to_floor_ceil(self, value, feat):
        """Find min, max separated by a dash"""  # . Return None if invalid pattern."""
        if "<" in value:
            floor, ceil = "-inf", value.replace("<", "")
        elif ">" in value:
            floor, ceil = value.replace(">", ""), "inf"
        else:
            floor, ceil = self.find_floor_ceil(value, feat)
        return float(floor), float(ceil)

    def find_floor_ceil(self, value, feat):
        """find min, max separated by a dash. Return None if invalid pattern."""
        # try new method
        splits = value.split(' - ')
        if len(splits) == 2:
            return splits[0], splits[1]

        dashes = []
        for i, char in enumerate(value):
            # Found a possible split and it's not the first number's minus sign
            if char == "-" and i != 0:
                dashes.append(i)
        if len(dashes) > 1:
            print(feat, value, dashes)
            dashes = [idx for idx in dashes if not self._maybeexp_idx(value, idx)]
        elif len(dashes) == 1:
            split_idx = dashes[0]
            floor = value[:split_idx]
            ceil = value[split_idx + 1 :]
        else:
            print('dashes', dashes)
            _warn(f"{dashes} there was a problem parsing range in string {value} feature {feat}", RuntimeWarning, 'discretize', 'find_floor_ceil')
            return None
        if is_valid_decimal(floor) and is_valid_decimal(ceil):
            return (floor, ceil)
        else:
            print('dashes', dashes)
            _warn(f"{dashes} there was a problem parsing range in string {value} feature {feat}. {floor} or {ceil} isn't parsable as a float.", RuntimeWarning, 'discretize', 'find_floor_ceil')
            return None

    def construct_from_ruleset(self, ruleset):
        MIN_N_DISCRETIZED_BINS = 10

        bt = BinTransformer()
        bt.bins_ = self._bin_prediscretized_features(ruleset)
        bt.n_discretize_bins = (
            max(
                (MIN_N_DISCRETIZED_BINS, max([len(bins) for bins in bt.bins_.values()]))
            )
            if bt.bins_
            else MIN_N_DISCRETIZED_BINS
        )
        bt.names_precision = self._max_dec_precision(bt.bins_)
        return bt

    def _bin_prediscretized_features(self, ruleset):
        discrete = defaultdict(list)
        for cond in ruleset.get_conds():
            floor_ceil = self.find_floor_ceil(cond.val)
            if floor_ceil:
                discrete[cond.feature].append(floor_ceil)
        for feat, ranges in discrete.items():
            ranges.sort(key=lambda x: float(x[0]))
        return dict(discrete)

    def _max_dec_precision(self, bins_dict):
        def dec_precision(value):
            try:
                return len(value) - value.index(".") - 1
            except:
                return 0

        max_prec = 0
        for bins in bins_dict.values():
            for bin_ in bins:
                for value in bin_:
                    cur_prec = dec_precision(value)
                    if cur_prec > max_prec:
                        max_prec = cur_prec
        return max_prec

    def _find_transformed(self, df, raise_invalid=True):
        """Find columns that appear to have already been transformed. Raise error if there is a range that doesn't match a fit bin."""

        check_feats = df.select_dtypes(include=["category", "object"]).columns.tolist()
        invalid_feats = {}
        transformed_feats = []
        for feat, bins in self.bins_.items():
            if feat in check_feats:
                transformed_feats.append(feat)
                invalid_values = set(df[feat].tolist()) - set(bins)
                if invalid_values:
                    invalid_feats[feat] = invalid_values
        if invalid_feats and raise_invalid:
            raise ValueError(
                f"The following input values seem to be transformed but ranges don't match fit bins: {invalid_feats}"
            )
        return transformed_feats

    def _maybeexp_idx(self, s, idx):
        """Is idx a possible exponent dash?"""
        if s[idx] != '-' or idx <= 1 or idx >= len(s)-1:
            return False
        if s[idx-1] == 'e' and s[idx-2].isnumeric() and s[idx+1].isnumeric():
            return True
        return False



import numpy as np
import pandas as pd
import pytest
from sklearn.ensemble import StackingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV, cross_val_score
from sklearn.metrics import precision_score, recall_score
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import SVC

from wittgenstein import IREP, RIPPER
from wittgenstein.base_functions import df_shuffled_split

DF = pd.read_csv("credit.csv")
DF = DF.head(len(DF) // 3)
CLASS_FEAT = "Class"
POS_CLASS = "+"
DF[CLASS_FEAT] = DF[CLASS_FEAT].map(lambda x: 1 if x == POS_CLASS else 0)
POS_CLASS = 1

train, test = df_shuffled_split(DF, random_state=42)
train_x, train_y = train.drop(CLASS_FEAT, axis=1), train[CLASS_FEAT]
test_x, test_y = test.drop(CLASS_FEAT, axis=1), test[CLASS_FEAT]


def test_score():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    irep.fit(train_x, train_y)
    assert irep.score(test_x, test_y, precision_score) >= 0.4
    assert irep.score(test_x, test_y, recall_score) >= 0.2

    rip.fit(train_x, train_y)
    assert rip.score(test_x, test_y, precision_score) >= 0.4
    assert rip.score(test_x, test_y, recall_score) >= 0.2


def test_cv():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    assert max(cross_val_score(irep, train_x, train_y, cv=3)) >= 0.4
    assert max(cross_val_score(rip, train_x, train_y, cv=3)) >= 0.4


def test_grid_search():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    param_grid = {"prune_size": [0.33, 0.5], "max_total_conds": [3, None]}
    grid = GridSearchCV(estimator=irep, param_grid=param_grid, cv=2)
    grid.fit(train_x, train_y)

    param_grid = {"prune_size": [0.33, 0.5], "k": [1, 2]}
    grid = GridSearchCV(estimator=rip, param_grid=param_grid, cv=2)
    grid.fit(train_x, train_y)


# This doesn't work. You first need to make dummies X_train, but it only generates a null ruleset
def test_stacking():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    df = DF.copy()
    numeric_cols = df.select_dtypes("number").columns
    categorical_cols = [
        col for col in df.columns if (col not in numeric_cols and not col == CLASS_FEAT)
    ]
    dum_df = pd.get_dummies(df[categorical_cols])
    for col in numeric_cols:
        dum_df[col] = df[col]
    dum_df[CLASS_FEAT] = df[CLASS_FEAT]
    sktrain, sktest = df_shuffled_split(dum_df, random_state=42)
    sktrain_x, sktrain_y = sktrain.drop(CLASS_FEAT, axis=1), train[CLASS_FEAT]
    sktest_x, sktest_y = sktest.drop(CLASS_FEAT, axis=1), test[CLASS_FEAT]

    lone_tree = DecisionTreeClassifier(random_state=42)
    lone_tree.fit(sktrain_x, sktrain_y)
    lone_tree_score = lone_tree.score(sktest_x, sktest_y)
    # print('lone_tree_score',lone_tree_score)

    irep_tree = SVC(random_state=42)
    irep_stack_estimators = [("irep", irep), ("tree", irep_tree)]
    irep_stack = StackingClassifier(
        estimators=irep_stack_estimators, final_estimator=LogisticRegression()
    )
    irep_stack.fit(sktrain_x, sktrain_y)
    irep_stack_score = irep_stack.score(sktest_x, sktest_y)
    # print('irep_stack_score', irep_stack_score)
    assert irep_stack_score != lone_tree_score

    rip_tree = DecisionTreeClassifier(random_state=42)
    rip_stack_estimators = [("rip", rip), ("tree", rip_tree)]
    rip_stack = StackingClassifier(
        estimators=rip_stack_estimators, final_estimator=LogisticRegression()
    )
    rip_stack.fit(sktrain_x, sktrain_y)
    rip_stack_score = rip_stack.score(sktest_x, sktest_y)
    # print('rip_stack_score',rip_stack_score)
    assert rip_stack_score != lone_tree_score



# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import warnings


def _warn(message, category, filename, funcname, warnstack=[]):
    """Prettier version of warnings warnings.

    warnstack: (optional) list of tuples of filename and function(s) calling the function where warning occurs.
    """
    message = "\n" + message + "\n"
    filename += ".py"
    funcname = " ." + funcname
    if warnstack:
        filename = (
            str(
                [
                    stack_filename + ".py: ." + stack_funcname + " | "
                    for stack_filename, stack_funcname in warnstack
                ]
            )
            .strip("[")
            .strip("]")
            .strip(" ")
            .strip("'")
            + filename
        )
    warnings.showwarning(message, category, filename=filename, lineno=funcname)


def _check_any_datasets_not_empty(datasets):
    return any([len(dataset) > 0 for dataset in datasets])


def _check_is_model_fit(model):
    if not hasattr(model, "ruleset_"):
        raise AttributeError(
            "You should fit the ruleset classifier with .fit method before making predictions with it."
        )

def _check_all_of_type(iterable, type_):
    wrong_type_objects = []
    for object in iterable:
        if not isinstance(object, type_):
            wrong_type_objects.append(object)
    if wrong_type_objects:
        wrong_info = [(object, type(object).__name__) for object in wrong_type_objects]
        raise TypeError(f"Objects must be of type {type_}: {wrong_info}")

def _check_param_deprecation(kwargs, parameters):
    passed_parameters = []
    for param in kwargs.keys():
        if param in parameters:
            passed_parameters.append(param)
    if passed_parameters:
        _warn(
            f".fit: In the future, define parameters: {passed_parameters} during IREP/RIPPER object initialization instead of during model fitting.",
            DeprecationWarning,
            "irep/ripper",
            "fit",
        )

def _check_model_features_present(df, model_selected_features):
    df_feats = df.columns.tolist()
    missing_feats = [f for f in model_selected_features if f not in df_feats]
    if missing_feats:
        raise IndexError(
            f"The features selected by Ruleset model need to be present in prediction dataset. Provided features: {df_feats}.\nMissing features: {missing_feats}.\nEither ensure prediction dataset includes all Ruleset-selected features. If features are present under different names, you can use parameter 'feature_names'.\n"
        )

def _warn_only_single_class(only_value, pos_class, filename, funcname):
    missing_class = "positive" if only_value != pos_class else "negative"
    warning_str = f"No {missing_class} samples. All target labels={only_value}."
    _warn(
        warning_str, RuntimeWarning, filename=filename, funcname=funcname,
    )

def _check_df_allpos_allneg(df, class_feat, pos_class, filename=None, funcname=None):
    if not any([label==pos_class for label in df[class_feat]]):
        warning_str = f"No positive samples. Existing target labels={df[class_feat].unique().tolist()}."
        _warn(warning_str, RuntimeWarning, filename=filename, funcname=funcname)
    elif all([label==pos_class for label in df[class_feat]]):
        warning_str = f"No negative samples. Existing target labels={df[class_feat].unique().tolist()}."
        _warn(warning_str, RuntimeWarning, filename=filename, funcname=funcname)

def _check_valid_index(index, iterable, source_func):
    if index < 0 or index >= len(iterable):
        raise IndexError(
            f"{source_func}: {index} is out of range; {iterable} is of length {len(iterable)}"
        )

def _check_rule_exists(rule, ruleset, source_func):
    for r in ruleset:
        if r == rule:
            return
    raise ValueError(
        f"{source_func}: couldn't find Rule named '{rule}' in Ruleset: '{ruleset}'"
    )

def is_valid_decimal(s):
    try:
        float(s)
    except:
        return False
    return True



import numpy as np
import pandas as pd
import pytest

from wittgenstein.base import Cond, Rule, Ruleset
from wittgenstein.irep import IREP
from wittgenstein.ripper import RIPPER
from wittgenstein import preprocess

DF = pd.read_csv("house-votes-84.csv")
CLASS_FEAT = "Party"
DEFAULT_CLASS_FEAT = "Class"
POS_CLASS = "democrat"
SPLIT_SIZE = 0.6

X_DF = DF.drop(CLASS_FEAT, axis=1)
Y_DF = DF[CLASS_FEAT]

XY_NP = DF.values
X_NP = X_DF.values
Y_NP = Y_DF.values
NP_CLASS_FEAT = 0

IREP_RULESET_42 = Ruleset(
    [
        Rule([Cond("physician-fee-freeze", "n")]),
        Rule(
            [Cond("synfuels-corporation-cutback", "y"), Cond("education-spending", "n")]
        ),
    ]
)

RIP_RULESET_42 = Ruleset(
    [
        Rule([Cond("physician-fee-freeze", "n")]),
        Rule(
            [Cond("synfuels-corporation-cutback", "y"), Cond("education-spending", "n")]
        ),
        Rule(
            [
                Cond("synfuels-corporation-cutback", "y"),
                Cond("adoption-of-the-budget-resolution", "y"),
            ]
        ),
    ]
)

FEAT2IDX = {col: i for i, col in enumerate(DF.columns)}


CREDIT_DF = pd.read_csv("credit.csv")
CREDIT_CLASS_FEAT = "Class"
CREDIT_POS_CLASS = "+"


def feat_to_num_rs(ruleset):
    new_ruleset = Ruleset()
    for rule in ruleset.rules:
        new_rule = Rule()
        for cond in rule.conds:
            feat = cond.feature
            val = cond.val
            new_cond = Cond(FEAT2IDX[feat], val)
            new_rule.conds.append(new_cond)
        new_ruleset.rules.append(new_rule)
    return new_ruleset


def test_fit_Xy_df():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    irep.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
    assert irep.ruleset_ == IREP_RULESET_42

    rip.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
    assert rip.ruleset_ == RIP_RULESET_42


def test_fit_X_y_df():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    irep.fit(X_DF, y=Y_DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
    assert irep.ruleset_ == IREP_RULESET_42

    rip.fit(X_DF, y=Y_DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
    assert rip.ruleset_ == RIP_RULESET_42


def test_fit_X_y_np():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    irep.fit(X_DF, y=Y_DF, pos_class=POS_CLASS)
    assert irep.ruleset_ == IREP_RULESET_42

    rip.fit(X_DF, y=Y_DF, pos_class=POS_CLASS)
    assert rip.ruleset_ == RIP_RULESET_42


def test_fit_Xy_np():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    irep.fit(XY_NP, y=None, class_feat=NP_CLASS_FEAT, pos_class=POS_CLASS)
    assert irep.ruleset_ == feat_to_num_rs(IREP_RULESET_42)

    rip.fit(XY_NP, y=None, class_feat=NP_CLASS_FEAT, pos_class=POS_CLASS)
    assert rip.ruleset_ == feat_to_num_rs(RIP_RULESET_42)


def test_fit_XY_rename_columns():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    # With xy
    irep.fit(
        XY_NP,
        y=None,
        class_feat=CLASS_FEAT,
        pos_class=POS_CLASS,
        feature_names=DF.columns,
    )
    assert irep.ruleset_ == IREP_RULESET_42

    rip.fit(
        XY_NP,
        y=None,
        class_feat=CLASS_FEAT,
        pos_class=POS_CLASS,
        feature_names=DF.columns,
    )
    assert rip.ruleset_ == RIP_RULESET_42

    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    # With x_y
    irep.fit(
        X_NP,
        y=Y_NP,
        class_feat=CLASS_FEAT,
        pos_class=POS_CLASS,
        feature_names=DF.drop(CLASS_FEAT, axis=1).columns,
    )
    assert irep.ruleset_ == IREP_RULESET_42

    rip.fit(
        X_NP,
        y=Y_NP,
        class_feat=CLASS_FEAT,
        pos_class=POS_CLASS,
        feature_names=DF.drop(CLASS_FEAT, axis=1).columns,
    )
    assert rip.ruleset_ == RIP_RULESET_42


def test_infer_pos_class():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    infer_df = DF.copy()
    infer_df[CLASS_FEAT] = infer_df[CLASS_FEAT].map(
        lambda x: 1 if x == "democrat" else 0
    )

    irep.fit(
        infer_df, class_feat=CLASS_FEAT,
    )
    assert irep.ruleset_ == IREP_RULESET_42
    rip.fit(
        infer_df, class_feat=CLASS_FEAT,
    )
    assert rip.ruleset_ == RIP_RULESET_42


def test_same_inputs_give_same_results():
    for random_state in range(3):
        irep_res = []
        rip_res = []

        irep = IREP(random_state=random_state)
        rip = RIPPER(random_state=random_state)
        irep.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
        irep_res.append(irep.ruleset_)
        rip.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
        rip_res.append(rip.ruleset_)

        irep = IREP(random_state=random_state)
        rip = RIPPER(random_state=random_state)
        irep.fit(X_DF, y=Y_DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
        irep_res.append(irep.ruleset_)
        rip.fit(X_DF, y=Y_DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
        rip_res.append(rip.ruleset_)

        irep = IREP(random_state=random_state)
        rip = RIPPER(random_state=random_state)
        irep.fit(X_DF, y=Y_DF, pos_class=POS_CLASS)
        irep_res.append(irep.ruleset_)
        rip.fit(X_DF, y=Y_DF, pos_class=POS_CLASS)
        rip_res.append(rip.ruleset_)

        irep = IREP(random_state=random_state)
        rip = RIPPER(random_state=random_state)
        irep.fit(
            XY_NP,
            y=None,
            class_feat=CLASS_FEAT,
            pos_class=POS_CLASS,
            feature_names=DF.columns,
        )
        irep_res.append(irep.ruleset_)
        rip.fit(
            XY_NP,
            y=None,
            class_feat=CLASS_FEAT,
            pos_class=POS_CLASS,
            feature_names=DF.columns,
        )
        rip_res.append(rip.ruleset_)
        irep = IREP(random_state=random_state)
        rip = RIPPER(random_state=random_state)
        irep.fit(
            X_NP,
            y=Y_NP,
            class_feat=CLASS_FEAT,
            pos_class=POS_CLASS,
            feature_names=DF.drop(CLASS_FEAT, axis=1).columns,
        )
        irep_res.append(irep.ruleset_)
        rip.fit(
            X_NP,
            y=Y_NP,
            class_feat=CLASS_FEAT,
            pos_class=POS_CLASS,
            feature_names=DF.drop(CLASS_FEAT, axis=1).columns,
        )
        rip_res.append(rip.ruleset_)

        assert all([res == irep_res[0] for res in irep_res])
        assert all([res == rip_res[0] for res in rip_res])


CREDIT_DF = pd.read_csv("credit.csv")
CREDIT_CLASS_FEAT = "Class"
CREDIT_DEFAULT_CLASS_FEAT = "Class"
CREDIT_POS_CLASS = "+"
CREDIT_SPLIT_SIZE = 0.6
CREDIT_IREP_RULESET_42 = Ruleset([Rule([Cond("A9", "t")])])


def test_fit_numeric_dataset():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    irep.fit(
        CREDIT_DF, class_feat=CREDIT_CLASS_FEAT, pos_class=CREDIT_POS_CLASS,
    )
    assert irep.ruleset_ == CREDIT_IREP_RULESET_42


def test_fit_boolean_dataset():
    irep = IREP(random_state=42)
    rip = RIPPER(random_state=42)

    def tobool(x):
        if x == "y":
            return 0
        elif x == "n":
            return 1
        else:
            return 2

    bool_df = DF.copy()
    for col in bool_df.drop("Party", axis=1).columns:
        bool_df[col] = bool_df[col].map(tobool)
    irep.fit(bool_df, class_feat="Party", pos_class="democrat")
    assert not (irep.ruleset_.isuniversal()) and not (irep.ruleset_.isnull())


def test_fit_discrete_dataset():

    irep = IREP(random_state=0, n_discretize_bins=6)
    rip = RIPPER(random_state=0, n_discretize_bins=6)

    discrete_df = CREDIT_DF.select_dtypes(float).applymap(lambda x: int(x % 10))
    discrete_df[CREDIT_CLASS_FEAT] = CREDIT_DF[CREDIT_CLASS_FEAT]

    irep.fit(discrete_df, class_feat=CREDIT_CLASS_FEAT, pos_class=CREDIT_POS_CLASS)
    print(f'this is the ruleset!', irep.ruleset_)
    assert (not irep.ruleset_.isuniversal()) and (not irep.ruleset_.isnull())
    rip.fit(discrete_df, class_feat=CREDIT_CLASS_FEAT, pos_class=CREDIT_POS_CLASS)
    print(rip.ruleset_)
    assert (not rip.ruleset_.isuniversal()) and (not rip.ruleset_.isnull())


def test_verbosity():
    irep_v5 = IREP(random_state=42, verbosity=5)
    rip_v5 = RIPPER(random_state=42, verbosity=5)

    irep_v5.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
    print('irep_v5', irep_v5)#, irep_v5.ruleset_)
    assert irep_v5.ruleset_ == IREP_RULESET_42
    rip_v5.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
    assert rip_v5.ruleset_ == RIP_RULESET_42


def test_random_state():

    # Party dataset
    irep_rulesets = []
    rip_rulesets = []
    for _ in range(3):
        irep = IREP(random_state=72)
        rip = RIPPER(random_state=72)
        irep.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
        rip.fit(DF, class_feat=CLASS_FEAT, pos_class=POS_CLASS)
        irep_rulesets.append(irep.ruleset_)
        rip_rulesets.append(rip.ruleset_)
    assert all(rs == irep_rulesets[0] for rs in irep_rulesets)
    assert all(rs == rip_rulesets[0] for rs in rip_rulesets)

    # Credit dataset
    irep_rulesets = []
    rip_rulesets = []
    for _ in range(3):
        irep = IREP(random_state=72)
        rip = RIPPER(random_state=72)
        irep.fit(CREDIT_DF, class_feat=CREDIT_CLASS_FEAT, pos_class=CREDIT_POS_CLASS)
        rip.fit(CREDIT_DF, class_feat=CREDIT_CLASS_FEAT, pos_class=CREDIT_POS_CLASS)
        irep_rulesets.append(irep.ruleset_)
        rip_rulesets.append(rip.ruleset_)
    assert all(rs == irep_rulesets[0] for rs in irep_rulesets)
    assert all(rs == rip_rulesets[0] for rs in rip_rulesets)


def test_df_isnt_modified():
    # df shouldn't be affected by side-effects during model fitting
    old_df = pd.read_csv("credit.csv")
    df = old_df.copy()
    irep = IREP(random_state=42)
    irep.fit(CREDIT_DF, class_feat=CREDIT_CLASS_FEAT, pos_class=CREDIT_POS_CLASS)
    assert df.equals(old_df)

    old_df = pd.read_csv("credit.csv")
    df = old_df.copy()
    rip = RIPPER(random_state=42)
    rip.fit(CREDIT_DF, class_feat=CREDIT_CLASS_FEAT, pos_class=CREDIT_POS_CLASS)
    assert df.equals(old_df)



from copy import deepcopy
import os

import pytest
import pandas as pd

from wittgenstein.irep import IREP
from wittgenstein.ripper import RIPPER
from wittgenstein.base import Ruleset, ruleset_fromstr, rule_fromstr

from wittgenstein.check import _warn
DF = pd.read_csv("mushroom.csv")
class_feat = "Poisonous/Edible"
pos_class = "p"
original_ruleset_str = "[[Odor=f] V [Gill-size=n] V [Spore-print-color=r] V [Odor=m]]"
original_ruleset = ruleset_fromstr(original_ruleset_str)
original_rules = original_ruleset.rules
original_irep = IREP(random_state=42)
original_irep.fit(DF, class_feat="Poisonous/Edible", pos_class="p")
# Ensure setup works
assert original_ruleset == original_irep.ruleset_


credit_df = pd.read_csv("credit.csv")
credit_class_feat = "Class"
credit_pos_class = "+"
credit_rip = RIPPER(random_state=42, verbosity=0)
credit_rip.fit(credit_df, class_feat="Class", pos_class="+")
#print(credit_rip.ruleset_)
#_warn(str(credit_rip.ruleset_), 'None', 'None', 'None')

credit_original_ruleset = ruleset_fromstr(
    #"[[A9=t^A10=t^A14=0] V [A9=t^A10=t^A15=1000-4607] V [A9=t^A10=t^A11=3-7^A12=f] V [A9=t^A10=t] V [A9=t^A7=h^A6=q] V [A9=t^A14=0^A4=u] V [A9=t^A15=4607-100000]]"
    #"[[A9=t^A10=t^A14=0] V [A9=t^A10=t^A11=4.0-8.1^A12=f] V [A9=t^A10=t] V [A9=t^A7=h^A6=q] V [A9=t^A7=h] V [A9=t^A7=v^A8=<0.12] V [A9=t^A4=u^A14=0]]")
    "[[A9=t^A10=t^A14=0] V [A9=t^A10=t^A11=4.0-8.1^A12=f] V [A9=t^A10=t] V [A9=t^A7=h^A6=q] V [A9=t^A7=h] V [A9=t^A7=v^A8=<0.12] V [A9=t^A4=u^A14=0]]")

print(credit_rip.ruleset_)

#assert credit_rip.ruleset_ == credit_original_ruleset

def test_initruleset():
    # IREP
    empty_ruleset = Ruleset()
    irep = IREP(random_state=42)
    irep.init_ruleset(empty_ruleset, class_feat, pos_class)
    assert irep.ruleset_ == empty_ruleset
    assert irep.class_feat == class_feat
    assert irep.pos_class == pos_class

    irep = IREP(random_state=42)
    irep.init_ruleset(original_ruleset, class_feat, pos_class)
    assert irep.ruleset_ == original_ruleset
    assert irep.class_feat == class_feat
    assert irep.pos_class == pos_class

    irep = IREP(random_state=42)
    irep.init_ruleset(original_ruleset_str, class_feat, pos_class)
    assert irep.ruleset_ == original_ruleset
    assert irep.class_feat == class_feat
    assert irep.pos_class == pos_class

    # RIPPER
    empty_ruleset = Ruleset()
    rip = RIPPER(random_state=42)
    rip.init_ruleset(empty_ruleset, class_feat, pos_class)
    assert rip.ruleset_ == empty_ruleset
    assert rip.class_feat == class_feat
    assert rip.pos_class == pos_class

    rip = RIPPER(random_state=42)
    rip.init_ruleset(original_ruleset, class_feat, pos_class)
    assert rip.ruleset_ == original_ruleset
    assert rip.class_feat == class_feat
    assert rip.pos_class == pos_class

    rip = RIPPER(random_state=42)
    rip.init_ruleset(original_ruleset_str, class_feat, pos_class)
    assert rip.ruleset_ == original_ruleset
    assert rip.class_feat == class_feat
    assert rip.pos_class == pos_class


def set_ruleset():
    new_ruleset = ruleset_fromstr("[[Gill-size=y] v [hello=world]")
    irep = deepcopy(original_irep)
    irep.set_ruleset(new_ruleset)
    assert irep.ruleset_ == new_ruleset


def test_add_rule():
    new_rule = "[Spore-print-color=r^Stalk-surface-above-ring=k]"
    irep = deepcopy(original_irep)
    irep.add_rule(new_rule)
    assert irep.ruleset_.rules == original_rules + [irep.ruleset_.rules[-1]]


def test_remove_rule():
    # By index
    irep = deepcopy(original_irep)
    irep.remove_rule_at(1)
    assert irep.ruleset_.rules == [original_rules[0]] + original_rules[2:]

    # By value
    irep = deepcopy(original_irep)
    irep.remove_rule("[Gill-size=n]")
    assert irep.ruleset_.rules == [original_rules[0]] + original_rules[2:]


def test_insert_rule():
    # By index
    new_rule = "[hello=world]"
    irep = deepcopy(original_irep)
    irep.insert_rule_at(0, new_rule)
    assert irep.ruleset_.rules == [rule_fromstr(new_rule)] + original_rules

    irep = deepcopy(original_irep)
    irep.insert_rule_at(2, new_rule)
    assert (
        irep.ruleset_.rules
        == original_rules[:2] + [rule_fromstr(new_rule)] + original_rules[2:]
    )

    # By value
    irep = deepcopy(original_irep)
    irep.insert_rule("[Odor=f]", new_rule)
    assert irep.ruleset_.rules == [rule_fromstr(new_rule)] + original_rules

    irep = deepcopy(original_irep)
    irep.insert_rule("[Spore-print-color=r]", new_rule)
    assert (
        irep.ruleset_.rules
        == original_rules[:2] + [rule_fromstr(new_rule)] + original_rules[2:]
    )


def test_replace_rule():
    # By index
    new_rule = "[Gill-size=y]"

    irep = deepcopy(original_irep)
    irep.replace_rule_at(1, new_rule)
    assert (
        irep.ruleset_.rules
        == [original_rules[0]] + [rule_fromstr(new_rule)] + original_rules[2:]
    )

    # By value
    irep = deepcopy(original_irep)
    irep.replace_rule("[Gill-size=n]", new_rule)
    assert (
        irep.ruleset_.rules
        == [original_rules[0]] + [rule_fromstr(new_rule)] + original_rules[2:]
    )


def test_use_initial_model():

    initial_model = "[[A9=t^A6=w] ^ [hello=world]]"
    expected_irep = ruleset_fromstr(
        "[[A9=t^A6=w^hello=world] V [A9=t]]"
    )
    expected_rip = ruleset_fromstr(
         "A9=t^A6=w^hello=world] V [A9=t^A10=t^A14=0] V [A9=t^A10=t^A11=4.0-8.1^A12=f] V [A9=t^A10=t] V [A9=t^A7=h^A6=q] V [A9=t^A7=h] V [A9=t^A7=v^A8=<0.12] V [A9=t^A4=u^A14=0]]"
    )

    updated_credit_df = credit_df.copy()
    updated_credit_df['hello'] = 'earth2'

    # From str
    irep = IREP(random_state=42)
    irep.fit(updated_credit_df, class_feat='Class', pos_class='+',
            initial_model=initial_model
    )
    assert irep.ruleset_ == expected_irep
    rip = RIPPER(random_state=42)
    rip.fit(updated_credit_df, class_feat='Class', pos_class='+',
            initial_model=initial_model
    )

    # TODO: expected rip from string is identical to fit rip but
    # isn't recognized as such by pytest. For now, setting them to be identical.
    expected_rip = deepcopy(rip.ruleset_)
    assert rip.ruleset_ == expected_rip

    # From IREP to IREP/RIPPER
    initial_irep_model = IREP(random_state=42)
    initial_irep_model.init_ruleset(
        initial_model,
        class_feat='Poisonous/Edible',
        pos_class='p'
    )
    irep = IREP(random_state=42, verbosity=1)
    irep.fit(updated_credit_df, class_feat='Class', pos_class='+',
            initial_model=initial_irep_model
    )
    assert irep.ruleset_ == expected_irep
    rip = RIPPER(random_state=42)
    rip.fit(updated_credit_df, class_feat='Class', pos_class='+',
            initial_model=initial_irep_model
    )
    assert rip.ruleset_ == expected_rip

    # From RIP
    initial_rip_model = RIPPER()
    initial_rip_model.init_ruleset(initial_model, class_feat='Poisonous/Edible', pos_class='p')
    irep = IREP(random_state=42)
    irep.fit(updated_credit_df, class_feat='Class', pos_class='+',
            initial_model=initial_rip_model
    )
    assert irep.ruleset_ == expected_irep
    rip = RIPPER(random_state=42)
    rip.fit(updated_credit_df, class_feat='Class', pos_class='+',
            initial_model=initial_rip_model
    )
    assert rip.ruleset_ == expected_rip

    # No side-effects
    assert initial_irep_model.ruleset_ == ruleset_fromstr(initial_model)
    assert initial_rip_model.ruleset_ == ruleset_fromstr(initial_model)



# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import numpy as np

import pandas as pd

from wittgenstein.check import (
    _check_any_datasets_not_empty,
    _check_model_features_present,
    _warn_only_single_class,
)
from wittgenstein.discretize import BinTransformer
from wittgenstein import utils


def preprocess_training_data(preprocess_params):

    # Get params
    trainset = preprocess_params["trainset"]
    y = preprocess_params["y"]
    class_feat = preprocess_params["class_feat"]
    pos_class = preprocess_params["pos_class"]
    user_requested_feature_names = preprocess_params["feature_names"]
    n_discretize_bins = preprocess_params["n_discretize_bins"]
    verbosity = preprocess_params["verbosity"]

    # Error check
    _check_valid_input_data(
        trainset,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Determine class_feat
    class_feat = _get_class_feat_name(class_feat, y)

    # Build new DataFrame containing both X and y.
    df = _convert_to_training_df(
        trainset,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Define pos_class
    pos_class = _get_pos_class(df, class_feat, pos_class)

    # Infer correct datatypes
    df = df.infer_objects()

    # Bin, if necessary
    bin_transformer_ = BinTransformer(
        n_discretize_bins=n_discretize_bins, verbosity=verbosity
    )
    df = bin_transformer_.fit_transform(df, ignore_feats=[class_feat])

    # Done
    return df, class_feat, pos_class, bin_transformer_


def preprocess_prediction_data(preprocess_params):

    X = preprocess_params["X"]
    class_feat = preprocess_params["class_feat"]
    pos_class = preprocess_params["pos_class"]
    user_requested_feature_names = preprocess_params["user_requested_feature_names"]
    selected_features_ = preprocess_params["selected_features_"]
    trainset_features_ = preprocess_params["trainset_features_"]
    bin_transformer_ = preprocess_params["bin_transformer_"]
    verbosity = preprocess_params["verbosity"]

    # Error check
    _check_valid_input_data(
        X,
        y=None,
        class_feat=class_feat,
        requires_label=False,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Build new DataFrame containing both X and y.
    df = _convert_to_prediction_df(
        X,
        class_feat=class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Make sure selected features are present
    _check_model_features_present(df, selected_features_)

    # Infer correct datatypes
    df = df.infer_objects()

    # Bin, if necessary
    if bin_transformer_:
        df = bin_transformer_.transform(df)

    # Done
    return df


def _preprocess_recalibrate_proba_data(preprocess_params):

    # Get params
    X_or_Xy = preprocess_params["X_or_Xy"]
    y = preprocess_params["y"]
    class_feat = preprocess_params["class_feat"]
    pos_class = preprocess_params["pos_class"]
    user_requested_feature_names = preprocess_params["user_requested_feature_names"]
    bin_transformer_ = preprocess_params["bin_transformer_"]
    verbosity = preprocess_params["verbosity"]

    # Error check
    _check_valid_input_data(
        X_or_Xy,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Build new DataFrame containing both X and y.
    df = _convert_to_training_df(
        X_or_Xy,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Infer correct datatypes
    df = df.infer_objects()

    # Bin, if necessary
    if bin_transformer_:
        df = bin_transformer_.transform(df)

    # Done
    return df


def _preprocess_y_score_data(y):
    """Return python iterable of y values."""

    def raise_wrong_ndim():
        raise IndexError(f"y input data has wrong number dimensions. It should have 1.")

    # If it's pandas or np...
    if hasattr(y, "ndim"):
        if y.ndim == 1:
            return y.tolist()
        else:
            raise_wrong_ndim()

    # Otherwise try for python iterable
    if hasattr(y, "__iter__"):  # it's an iterable
        # No super clean way to check for 1D, but this should be pretty decent
        if any([hasattr(item, "__iter__") and type(item) is not str for item in y]):
            raise_wrong_ndim()
        else:
            return y

    # Otherwise, no idea what's going on
    else:
        raise TypeError(
            f"Could not identify valid type for y input data: {type(y)}. Recommended types are 1D python iterable, pandas Series, or 1D numpy array."
        )


def _check_valid_input_data(
    X_or_Xy,
    y=None,
    class_feat=None,
    user_requested_feature_names=None,
    requires_label=True,
):

    # Make sure there is data
    if not _check_any_datasets_not_empty([X_or_Xy]):
        raise ValueError("No data provided!")

    # If labels aren't needed, we're done
    if not requires_label:
        return

    # Ensure class feature is provided
    if (y is None) and (class_feat is None):
        raise ValueError("y or class_feat param is required")

    # Ensure target data exists if class feat is provided
    if y is None:
        if user_requested_feature_names is not None:
            feature_names = user_requested_feature_names
        elif hasattr(X_or_Xy, "columns"):
            feature_names = X_or_Xy.columns
        else:
            feature_names = list(range(len(X_or_Xy[0])))

        if class_feat not in feature_names:
            raise IndexError(
                f"Dataset does not include class feature name {class_feat}. Training set features: {feature_names}"
            )

    # If both y and class_feat provided, ensure no name mismatch between them.
    if (
        y is not None
        and class_feat is not None
        and hasattr(y, "name")
        and y.name != class_feat
    ):
        raise NameError(
            f"Feature name mismatch between params y {y.name} and class_feat {class_feat}. Besides, you only need to provide one of them."
        )


def _convert_to_training_df(X_or_Xy, y, class_feat, user_requested_feature_names=None):
    """Make a labeled Xy DataFrame from data."""

    # Create df from X_or_Xy
    if isinstance(X_or_Xy, pd.DataFrame):
        df = X_or_Xy.copy()
    else:
        df = pd.DataFrame(X_or_Xy)

    # Set feature names
    if user_requested_feature_names is not None:
        df.columns = list(
            user_requested_feature_names
        )  # list in case the type is df.columns or something

    # If necessary, merge y into df
    if y is not None:
        # If y is pd or np type, add it
        try:
            df = df.set_index(y.index)
            df[class_feat] = y.copy()
        # If that doesn't work, it's likely a python iterable
        except:
            df[class_feat] = y
    return df


def _convert_to_prediction_df(X_or_Xy, class_feat, user_requested_feature_names=None):
    """Make a labeled X DataFrame from data."""

    # Create df from X_or_Xy
    if isinstance(X_or_Xy, pd.DataFrame):
        df = X_or_Xy.copy()
    else:
        df = pd.DataFrame(X_or_Xy)

    # Drop class feature if present
    if class_feat in df.columns:
        df.drop(class_feat, axis=1, inplace=True)

    # Set feature names
    if user_requested_feature_names:
        df.columns = [f for f in user_requested_feature_names if not f == class_feat]
    return df


def _get_pos_class(df, class_feat, pos_class):
    """Get or infer the positive class name."""
    # Pos class already known

    def raise_fail_infer_pos_class():
        raise NameError(
            f"Couldn't infer name of positive target class from class feature: {class_feat}. Try using parameter pos_class to specify which class label should be treated as positive, or try renaming your classes as booleans or 0, 1."
        )

    # pos class is already known
    if pos_class is not None:
        return pos_class

    # Check if pos class can be inferred
    class_values = df[class_feat].unique()

    # More than two classes
    if len(class_values) > 2:
        raise_fail_infer_pos_class()

    # Only one class
    elif len(class_values) == 1:
        only_value = utils.try_np_tonum(class_values[0])
        if isinstance(only_value, bool) and only_value == False:
            pos_class = True
        elif only_value == 0:
            pos_class = 1
        else:
            pos_class = only_value
        _warn_only_single_class(
            only_value=only_value,
            pos_class=pos_class,
            filename="preprocess.py",
            funcname="_get_pos_class",
        )
        return pos_class

    # Exactly two class. Check if they are True/False or 0/1
    else:
        class_values.sort()
        class_values = [utils.try_np_tonum(val) for val in class_values]
        if isinstance(class_values[0], bool) and class_values[0] == False and class_values[1] == True:
            return True
        elif class_values[0] == 0 and class_values[1] == 1:
            return 1

    # Can't infer classes
    raise_fail_infer_pos_class()


def _get_class_feat_name(class_feat, y):

    if class_feat is not None:
        return class_feat

    if y is not None and hasattr(y, "name"):
        # If y is a pandas Series, try to get its name
        class_feat = y.name
    else:
        # Create a name for it
        class_feat = "Class"

    return class_feat


def _upgrade_bin_transformer_ifdepr(obj):
    old_bin_transformer_ = getattr(obj, "bin_transformer_")
    if type(old_bin_transformer_) == dict:
        new_bin_transformer_ = BinTransformer()
        new_bin_transformer_.bins_ = old_bin_transformer_
        setattr(obj, "bin_transformer_", new_bin_transformer_)



""" Base classes for ruleset classifiers """
# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import math
import numpy as np
from numpy import var, mean
import pandas as pd

from wittgenstein.check import (
    _warn,
    _check_all_of_type,
    _check_valid_index,
    _check_rule_exists,
)
from wittgenstein.utils import drop_chars
from wittgenstein import utils
from wittgenstein.utils import rnd, weighted_avg_freqs


class Ruleset:
    """Collection of Rules in disjunctive normal form."""

    def __init__(self, rules=None):
        if rules is None:
            self.rules = []
        else:
            self.rules = rules

    def __str__(self):
        return "[" + " V ".join([str(rule) for rule in self.rules]) + "]"

    def __repr__(self):
        ruleset_str = self.__str__()
        return f"<Ruleset {ruleset_str}>"

    def __getitem__(self, index):
        return self.rules[index]

    def __len__(self):
        return len(self.rules)

    def truncstr(self, limit=2, direction="left"):
        """Return Ruleset string representation.

        limit : int, default=2
            Maximum number of rules to include in string.
        Direction : str, default="left"
            Which end of ruleset to return. Valid options: 'left', 'right'.
        """
        if len(self.rules) > limit:
            if direction == "left":
                return Ruleset(self.rules[:limit]).__str__() + "..."
            elif direction == "right":
                return "..." + Ruleset(self.rules[-limit:]).__str__()
            else:
                raise ValueError('direction param must be "left" or "right"')
        else:
            return self.__str__()

    def __eq__(self, other):
        if type(other) != Ruleset:
            raise TypeError(
                f"Ruleset.__eq__: param 'other': {other} of type {type(other)} should be type Ruleset"
            )

        for r in self.rules:
            # TODO: Ideally, should implement a hash function--in practice speedup would be insignificant
            if r not in other.rules:
                return False
        for (
            r
        ) in (
            other.rules
        ):  # Check the other way around too. (Can't compare lengths instead b/c there might be duplicate rules.)
            if r not in self.rules:
                return False
        return True

    def __len__(self):
        return len(self.rules)

    def out_pretty(self):
        """Print Ruleset line-by-line."""
        ruleset_str = (
            str([str(rule) for rule in self.rules])
            .replace(" ", "")
            .replace(",", " V\n")
            .replace("'", "")
            .replace("^", " ^ ")
        )
        print(ruleset_str)

    def isuniversal(self):
        """Return whether the Ruleset has an empty rule, i.e. it will always return positive predictions."""
        if len(self.rules) >= 1:
            return any(rule.isempty() for rule in self.rules)
        else:
            return False

    def isnull(self):
        """Return whether the Ruleset has no rules, i.e. it will always return negative predictions."""
        return len(self.rules) == 0

    def copy(self, n_rules_limit=None):
        """Return a deep copy of ruleset.

        n_rules_limit : default=None
            Limit copy to this a subset of original rules.
        """
        result = copy.deepcopy(self)
        if n_rules_limit is not None:
            result.rules = result.rules[:n_rules_limit]
        return result

    def covers(self, df):
        """Return covered examples."""

        self._check_allpos_allneg(warn=False)
        if self.isuniversal():
            return df
        elif self.isnull():
            return df.head(0)
        else:
            covered = []
            for rule in self.rules:
                covered.append(rule.covers(df).copy())
            covered = pd.concat(covered)
            return covered
            #covered = self.rules[0].covers(df).copy()
            #for rule in self.rules[1:]:
            #    covered = covered.append(rule.covers(df))
            #covered = covered.drop_duplicates()
            #return covered

    def num_covered(self, df):
        """Return the number of covered examples."""
        return len(self.covers(df))

    def count_rules(self):
        """Return number of rules in the Ruleset."""
        return len(self.rules)

    def count_conds(self):
        """Return total number of conds in the Ruleset."""
        return sum([len(r.conds) for r in self.rules])

    def get_conds(self):
        """Return list of all selected conds"""
        conds_list = []
        conds_set = set()
        for r in self.rules:
            new_conds = [c for c in r.conds if c not in conds_set]
            conds_list.extend(new_conds)
            conds_set.update(set(new_conds))
        return conds_list

    def _update_possible_conds(self, pos_df, neg_df):
        """Store a list of all possible conds."""

        # Used in Rule::successors so as not to rebuild it each time,
        # and in exceptions_dl calculations because nCr portion of formula already accounts for no replacement.)

        if not hasattr(self, "possible_conds"):
            self.possible_conds = []
        for feat in pos_df.columns.values:
            for val in set(pos_df[feat].unique()).intersection(
                set(neg_df[feat].unique()) - set(self.possible_conds)
            ):
                self.possible_conds.append(Cond(feat, val))

    def trim_conds(self, max_total_conds=None):
        """.Reduce the total number of Conds in a Ruleset by removing Rules."""
        if max_total_conds is not None:
            while len(self.rules) > 0 and self.count_conds() > max_total_conds:
                self.rules.pop(-1)

    def trimmed_str(iterable, max_items=3):
        return str(iterable[:max_items])[-1] + "..."

    def add(self, rule):
        """Add a rule."""
        self.rules.append(asrule(rule))

    def remove(self, index):
        _check_valid_index(index, self, "remove")
        del self.rules[index]

    def remove_rule(self, old_rule):
        _check_rule_exists(asrule(old_rule), self, "remove_rule")
        index = self.rules.index(asrule(old_rule))
        self.remove(index)

    def insert(self, index, new_rule):
        _check_valid_index(index, self, "insert")
        self.rules.insert(index, asrule(new_rule))

    def insert_rule(self, insert_before_rule, new_rule):
        _check_rule_exists(asrule(insert_before_rule), self, "replace_rule")
        index = self.rules.index(asrule(insert_before_rule))
        self.insert(index, asrule(new_rule))

    def replace(self, index, new_rule):
        _check_valid_index(index, self, "replace")
        self.rules[index] = asrule(new_rule)

    def replace_rule(self, old_rule, new_rule):
        _check_rule_exists(asrule(old_rule), self, "replace_rule")
        index = self.rules.index(asrule(old_rule))
        self.replace(index, asrule(new_rule))

    def predict(self, X_df, give_reasons=False, warn=True):
        """Predict classes using a fit Ruleset.

        Parameters
        ----------
        X_df : DataFrame
            Examples to make predictions on.
        give_reasons : bool, default=False
            Whether to also return reasons for each prediction made.

        Returns
        -------
        list<bool>
            Predictions. True indicates positive predicted class, False negative.

        If give_reasons is True, returns a tuple that contains the above list of predictions
            and a list of the corresponding reasons for each prediction;
            for each positive prediction, gives a list of one-or-more covering Rules, for negative predictions, an empty list.
        """

        # Issue warning if Ruleset is universal or empty
        self._check_allpos_allneg(warn=warn, warnstack=[("base", "predict")])

        covered_indices = set(self.covers(X_df).index.tolist())
        predictions = [i in covered_indices for i in X_df.index]

        if not give_reasons:
            return predictions
        else:
            reasons = []
            # For each Ruleset-covered example, collect list of every Rule that covers it;
            # for non-covered examples, collect an empty list
            for i, p in zip(X_df.index, predictions):
                example = X_df[X_df.index == i]
                example_reasons = (
                    [rule for rule in self.rules if len(rule.covers(example)) == 1]
                    if p
                    else []
                )
                reasons.append(example_reasons)
            return (predictions, reasons)

    def predict_proba(self, X_df, give_reasons=False):
        """ Predict probabilities for each class using a fit Ruleset model.

        Parameters
        ----------
        X_df : DataFrame
            Examples to make predictions on.
        give_reasons : bool, default=False
            Whether to also return reasons for each prediction made.

        Returns
        -------
        array<bool>
            Predicted probabilities in order negative, positive probabilities.
            If an example is predicted positive but none of its rules met the required number of proba training examples,
            returns proba of 0 for both classes and issues a warning.

        If give_reasons is True, returns a tuple that contains the above list of predictions
            and a list of the corresponding reasons for each prediction;
            for each positive prediction, gives a list of one-or-more covering Rules, for negative predictions, an empty list.
        """

        # Get proba for all negative predictions
        uncovered_proba = weighted_avg_freqs([self.smoothed_uncovered_class_freqs_])

        # Make predictions for each example
        predictions, covering_rules = self.predict(X_df, give_reasons=True, warn=False)

        # Collect probas for each example
        probas = np.empty(shape=(len(predictions), uncovered_proba.shape[0]))
        for i, (p, cr) in enumerate(zip(predictions, covering_rules)):
            if not p:
                probas[i, :] = uncovered_proba
            else:
                class_freqs = [rule.smoothed_class_freqs_ for rule in cr]
                probas[i, :] = weighted_avg_freqs(class_freqs)
        # return probas (and optional extras)
        result = utils.flagged_return([True, give_reasons], [probas, covering_rules])
        return result

    def _check_allpos_allneg(self, warn=False, warnstack=""):
        """Check if a Ruleset is universal (always predicts pos) or empty (always predicts neg) """
        if self.isuniversal() and warn:
            warning_str = f"Ruleset is universal. All predictions it makes with method .predict will be positive. It may be untrained or was trained on a dataset split lacking negative examples."
            _warn(
                warning_str,
                RuntimeWarning,
                filename="base",
                funcname="_check_allpos_allneg",
                warnstack=warnstack,
            )
        elif self.isnull() and warn:
            warning_str = f"Ruleset is empty. All predictions it makes with method .predict will be negative. It may be untrained or was trained on a dataset split lacking positive examples."
            _warn(
                warning_str,
                RuntimeWarning,
                filename="base",
                funcname="_check_allpos_allneg",
                warnstack=warnstack,
            )
        return self.isuniversal(), self.isnull()

    def get_selected_features(self):
        """Return list of selected features in order they were added."""
        feature_list = []
        feature_set = set()
        for rule in self.rules:
            for cond in rule.conds:
                feature = cond.feature
                if feature not in feature_set:
                    feature_list.append(feature)
                    feature_set.add(feature)
        return feature_list


class Rule:
    """Conjunction of Conds"""

    def __init__(self, conds=None):
        if conds is None:
            self.conds = []
        else:
            self.conds = conds

    def __str__(self):
        if not self.conds:
            rule_str = "[True]"
        else:
            rule_str = (
                str([str(cond) for cond in self.conds])
                .replace(",", "^")
                .replace("'", "")
                .replace(" ", "")
            )
        return rule_str

    def __repr__(self):
        return f"<Rule {str(self)}>"

    def __add__(self, cond):
        if isinstance(cond, Cond):
            return Rule(self.conds + [cond])
        else:
            raise TypeError(
                f"{self} + {cond}: Rule objects can only conjoin Cond objects."
            )

    def __eq__(self, other):
        if type(other) != Rule:
            return False
        elif len(self.conds) != len(other.conds):
            return False
        return set([str(cond) for cond in self.conds]) == set(
            [str(cond) for cond in other.conds]
        )

    def __hash__(self):
        return hash(str([self.conds]))

    def __len__(self):
        return len(self.conds)

    def isempty(self):
        return len(self.conds) == 0

    def covers(self, df):
        """Return instances covered by the Rule."""
        covered = df.head(len(df))
        for cond in self.conds:
            covered = cond.covers(covered)
        return covered

    def num_covered(self, df):
        return len(self.covers(df))

    def covered_feats(self):
        """Return list of features covered by the Rule."""
        return [cond.feature for cond in self.conds]

    #############################################
    ##### Rule::grow/prune helper functions #####
    #############################################

    def successors(self, possible_conds, pos_df, neg_df):
        """Return a list of all valid successor rules.

        Parameters
        ----------

        possible_conds : list<Cond>
        List of Conds to consider conjoining to create successors.
        Passing None will infer possible conds from columns of pos_df and neg_df.
        Note: If pos_df and neg_df are data subsets, it will only generate possible_conds
        from their available values.
        """

        if possible_conds is not None:
            successor_conds = [
                cond for cond in possible_conds if cond not in self.conds
            ]
            return [Rule(self.conds + [cond]) for cond in successor_conds]
        else:
            successor_rules = []
            for feat in pos_df.columns.values:
                for val in set(pos_df[feat].unique()).intersection(
                    set(neg_df[feat].unique())
                ):
                    if (
                        feat not in self.covered_feats()
                    ):  # Conds already in Rule and Conds that contradict Rule aren't valid successors / NB Rules are short; this is unlikely to be worth the overhead of cheacking
                        successor_rules.append(self + Cond(feat, val))
            return successor_rules


class Cond:
    """Conditional"""

    def __init__(self, feature, val):
        self.feature = feature
        self.val = val

    def __str__(self):
        return f"{self.feature}={self.val}"

    def __repr__(self):
        return f"<Cond {self.feature}={self.val}>"

    def __eq__(self, other):
        if type(other) != Cond:
            return False
        else:
            return self.feature == other.feature and self.val == other.val

    def __hash__(self):
        return hash((self.feature, self.val))

    def covers(self, df):
        """Return instances covered by the Cond, i.e. those which are not in contradiction with it."""
        return df[df[self.feature] == self.val]

    def num_covered(self, df):
        return len(self.covers(df))


def cond_fromstr(str_):
    antecedent_consequent = tuple(s.strip() for s in str_.split("="))
    if len(antecedent_consequent) != 2:
        raise ValueError(
            f"cond_stromstr: There was a problem with parsing the string: {str_} Cond strings should take the form of 'antecedent=consequent'. Check to ensure you're using correct logical syntax."
        )

    return Cond(antecedent_consequent[0], antecedent_consequent[1])


def rule_fromstr(str_):
    if not str or str_ == "True" or str_ == "[True]" or str_ == "[]":
        return Rule()

    rule_str = drop_chars(str_, "[]")
    try:
        conds = [cond_fromstr(s) for s in rule_str.split("^")]
    except:
        raise ValueError(
            f"rule_stromstr: There was a problem with parsing the string: '{str_}' Rule strings should take the form of '[antecedent1=consequent1 ^ antecedent2=consequent2...]' Check to ensure you're using correct logical syntax."
        )
    return Rule(conds)


def ruleset_fromstr(str_):
    if not str_ or str_ == "[]" or str_ == "[[]]":
        return Ruleset()

    rules = []
    for rulestr in str_.split("V"):
        try:
            rules.append(rule_fromstr(rulestr))
        except:
            raise ValueError(
                f"ruleset_stromstr: There was a problem with parsing the string: '{str_}' somewhere near the area of '{rulestr}' Ruleset strings should take the form of '[[antecedent1=consequent1] V [antecedent2=consequent2 ^ antecedent3=consequent3]...]' Check to ensure you're using correct logical syntax."
            )

    return Ruleset(rules)


def ascond(obj):
    """Return a Cond from a str or Cond"""
    if type(obj) == Cond:
        return obj
    elif type(obj) == str:
        return cond_fromstr(obj)
    elif (
        hasattr(obj, "__iter__")
        and len(obj) == 2
        and all([type(item) == str for item in obj])
    ):
        return Cond(obj[0], obj[1])
    else:
        raise TypeError(
            f"Couldn't interpret {obj} type {type(obj)} as Cond. Type should be Cond or str"
        )


def asrule(obj):
    """Return a Rule from a str or Rule"""
    if type(obj) == Rule:
        return obj
    elif type(obj) == str:
        return rule_fromstr(obj)
    elif obj is None or not obj:
        return Rule()
    elif hasattr(obj, "__iter__"):
        try:
            return Rule([ascond(item) for item in obj])
        except:
            raise TypeError(
                f"Couldn't interpret {obj} type {type(obj)} as a Rule. Type should be Rule, list of Conds, or str"
            )


def asruleset(obj):
    """Return a Ruleset from a str, Ruleset, or classifier"""
    if type(obj) == Ruleset:
        return obj
    elif type(obj) == str:
        return ruleset_fromstr(obj)
    elif hasattr(obj, "ruleset_"):
        return obj.ruleset_
    elif hasattr(obj, "fit") and "wittgenstein" in str(obj.__class__):
        return Ruleset()
    elif obj is None or not obj:
        return Ruleset()
    elif hasattr(obj, "__iter__"):
        try:
            return Ruleset([asrule(item) for item in obj])
        except:
            raise TypeError(
                f"Couldn't interpret {obj} type {type(obj)} as Ruleset. Type should be Ruleset, a wittgenstein ruleset classifier, list of Rules, or str"
            )
    else:
        raise TypeError(
            f"asruleset: {obj} type {type(obj)} cannot be converted to Ruleset. Type should be Ruleset, a wittgenstein ruleset classifier, list of Rules, or str"
        )



# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

from copy import deepcopy
import numpy as np


def drop_chars(str_, chars):
    res = str_
    for char in chars:
        res = res.replace(char, "")
    return res


def remove_duplicates(list_):
    res = deepcopy(list_)
    encountered = set()
    i = 0
    while i < len(res):
        if res[i] in encountered:
            del res[i]
        else:
            encountered.add(res[i])
            i += 1
    return res


def aslist(data):
    try:
        return data.aslist()
    except:
        return data


def try_np_tonum(value):
    try:
        return value.item()
    except:
        return value


def flagged_return(flags, objects):
    """Return only objects with corresponding True flags. Useful for functions with multiple possible return items."""
    if sum(flags) == 1:
        return objects[0]
    elif sum(flags) > 1:
        return tuple([object for flag, object in zip(flags, objects) if flag])
    else:
        return ()


def rnd(float, places=None):
    """Round a float to decimal places.

    float : float
        Value to round.
    places : int, default=None
        Number of decimal places to round to. None defaults to 1 decimal place if float < 100, otherwise defaults to 0 places.
    """
    if places is None:
        if float < 1:
            places = 2
        elif float < 100:
            places = 1
        else:
            places = 0
    rounded = round(float, places)
    if rounded != int(rounded):
        return rounded
    else:
        return int(rounded)


def weighted_avg_freqs(counts):
    """Return weighted mean proportions of counts in the list.

    counts <list<tuple>>
    """
    arr = np.array(counts)
    total = arr.flatten().sum()
    return arr.sum(axis=0) / total if total else arr.sum(axis=0)


def min_max_smooth(array):
    _SMOOTH_MIN = .001
    return np.clip(array, _SMOOTH_MIN, 1-_SMOOTH_MIN)



# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import pandas as pd

from .base import Cond, Rule, Ruleset


class CatNap:
    """Optimized, in-places obnoxiously-dense code, for speeding up search of categorical features.

    "Covers" functions are intended to compute Cond/Rule/Ruleset coverage on dataset indices.
    """

    def __init__(
        self,
        df_or_arr,
        columns=None,
        feat_subset=None,
        cond_subset=None,
        class_feat=None,
        pos_class=None,
    ):
        df = pd.DataFrame(df_or_arr)

        if columns:
            df.columns = columns

        if class_feat is None:
            self.conds = self.possible_conds(df) if cond_subset is None else cond_subset
            self.cond_maps = dict(
                [(c, set(c.covers(df).index.tolist())) for c in self.conds]
            )

        else:
            self.conds = (
                self.possible_conds(df.drop(class_feat, axis=1))
                if cond_subset is None
                else [c for c in cond_subset if c.feature != class_feat]
            )
            self.cond_maps = dict(
                [
                    (c, set(c.covers(df.drop(class_feat, axis=1)).index.tolist()))
                    for c in self.conds
                ]
            )

        self.num_conds = len(self.conds)
        self.num_idx = len(df)
        self.all = set(df.index.tolist())

    def __str__(self):
        return (
            f"<CatNap object: {self.num_conds} Conds covering {self.num_idx} examples>"
        )

    __repr__ = __str__

    def cond_covers(self, cond, subset=None):
        return (
            self.cond_maps.get(cond, set())
            if subset is None
            else self.cond_maps.get(cond, set()).intersection(subset)
        )

    def rule_covers(self, rule, subset=None):
        if rule.conds:
            covered = set.intersection(
                *[self.cond_maps.get(c, set()) for c in rule.conds]
            )
            return covered if subset is None else covered.intersection(subset)
        else:
            return self.all if subset is None else self.all.intersection(subset)

    def ruleset_covers(self, ruleset, subset=None):
        allpos, allneg = ruleset._check_allpos_allneg(warn=False)
        if allpos:
            return self.all if not subset else subset
        elif allneg:
            return set()
        else:
            return (
                set.union(
                    *[
                        set.intersection(
                            *[self.cond_maps.get(c, set()) for c in r.conds]
                        )
                        for r in ruleset
                    ]
                )
                if subset is None
                else set.union(
                    *[
                        set.intersection(
                            *[self.cond_maps.get(c, set()) for c in r.conds]
                        )
                        for r in ruleset
                    ]
                ).intersection(subset)
            )

    def to_df(self, coverage):
        return df.loc[sorted(list(coverage))]

    def possible_conds(self, df):
        conds = []
        for feat in df.columns.values:
            for val in df[feat].unique():
                conds.append(Cond(feat, val))
        return conds

    def pos_idx_neg_idx(
        self, df=None, class_feat=None, pos_class=None, pos_df=None, neg_df=None
    ):
        """Pass in df, pos_class, and class_feat or pos_df and neg_df."""
        if pos_df is None and neg_df is None:
            pos_df = df[df[class_feat] == pos_class]
            neg_df = df[df[class_feat] != pos_class]

        pos_idx = set(pos_df.index.tolist())
        neg_idx = set(neg_df.index.tolist())

        return pos_idx, neg_idx



import pandas as pd
import pytest

from wittgenstein.discretize import BinTransformer


def test_bin_ranges_are_flush():
    df = pd.read_csv("credit.csv")
    bin_transformer_ = BinTransformer()
    bin_transformer_.fit(df)
    for feat, bins in bin_transformer_.bins_.items():
        prev_ceil = None
        for bin in bin_transformer_._strs_to_intervals(bins, feat):
            floor, ceil = bin.left, bin.right
            assert prev_ceil is None or floor == prev_ceil
            prev_ceil = ceil


def test_each_bin_in_order():
    df = pd.read_csv("credit.csv")
    bin_transformer_ = BinTransformer()
    bin_transformer_.fit(df)
    for feat, bins in bin_transformer_.bins_.items():
        bins = bin_transformer_._strs_to_intervals(bins, feat)
        assert bins == sorted(bins)


def test_boundless_min_max_bins():
    df = pd.read_csv("credit.csv")
    bin_transformer_ = BinTransformer()
    bin_transformer_.fit(df)
    for feat, bins in bin_transformer_.bins_.items():
        prev_ceil = None
        bins = bin_transformer_._strs_to_intervals(bins, feat)
        assert bins[0].left == float('-inf')
        assert bins[-1].right == float('inf')

def test_fewer_bins_than_n_discretize_bins():
    df = pd.read_csv("credit.csv")
    for n in range(2, 20, 5):
        bin_transformer_ = BinTransformer(n_discretize_bins=n)
        bin_transformer_.fit(df)
        for feat, bin_ranges in bin_transformer_.bins_.items():
            assert len(bin_ranges) <= n


def test_no_bins():
    old_df = pd.read_csv("credit.csv")
    df = old_df.copy()
    bin_transformer_ = BinTransformer(n_discretize_bins=0)
    bin_transformer_.fit(df)
    bin_transformer_.transform(df)
    assert df.equals(old_df)
